package routingtables

import (
	"context"
	"fmt"

	"code.fbi.h-da.de/danet/gosdn/controller/customerrs"
	query "code.fbi.h-da.de/danet/gosdn/controller/store"

	"github.com/google/uuid"
	"go.mongodb.org/mongo-driver/bson"
	"go.mongodb.org/mongo-driver/bson/primitive"
	"go.mongodb.org/mongo-driver/mongo"
	"go.mongodb.org/mongo-driver/mongo/options"
)

const storeName = "routing-table-store.json"

// Store defines a RoutingTable store interface.
type Store interface {
	Add(context.Context, RoutingTable) error
	Update(context.Context, RoutingTable) error
	Delete(context.Context, RoutingTable) error
	Get(context.Context, query.Query) (RoutingTable, error)
	GetAll(context.Context) ([]RoutingTable, error)
}

// DatabaseRoutingTableStore is a database store for routingTables.
type DatabaseRoutingTableStore struct {
	collection *mongo.Collection
}

// NewDatabaseRoutingTableStore returns a RoutingTableStore.
func NewDatabaseRoutingTableStore(db *mongo.Database) Store {
	collection := db.Collection(storeName)

	return &DatabaseRoutingTableStore{
		collection: collection,
	}
}

// Get takes a routing-tables's UUID or name and returns the entries.
func (s *DatabaseRoutingTableStore) Get(ctx context.Context, query query.Query) (RoutingTable, error) {
	var loadedRoutingTable RoutingTable

	if query.ID.String() != "" {
		loadedRoutingTable, err := s.getByID(ctx, query.ID)
		if err != nil {
			return loadedRoutingTable, customerrs.CouldNotFindError{ID: query.ID, Name: query.Name}
		}

		return loadedRoutingTable, nil
	}

	loadedRoutingTable, err := s.getByName(ctx, query.Name)
	if err != nil {
		return loadedRoutingTable, customerrs.CouldNotFindError{ID: query.ID, Name: query.Name}
	}

	return loadedRoutingTable, nil
}

func (s *DatabaseRoutingTableStore) getByID(ctx context.Context, idOfRoutingTable uuid.UUID) (routingTable RoutingTable, err error) {
	result := s.collection.FindOne(ctx, bson.D{primitive.E{Key: "_id", Value: idOfRoutingTable.String()}})
	if result == nil {
		return routingTable, customerrs.CouldNotFindError{ID: idOfRoutingTable}
	}

	err = result.Decode(&routingTable)
	if err != nil {
		return routingTable, customerrs.CouldNotMarshallError{Identifier: idOfRoutingTable, Type: routingTable, Err: err}
	}

	return routingTable, nil
}

func (s *DatabaseRoutingTableStore) getByName(ctx context.Context, nameOfRoutingTable string) (loadedRoutingTable RoutingTable, err error) {
	result := s.collection.FindOne(ctx, bson.D{primitive.E{Key: "name", Value: nameOfRoutingTable}})
	if result == nil {
		return loadedRoutingTable, customerrs.CouldNotFindError{Name: nameOfRoutingTable}
	}

	err = result.Decode(&loadedRoutingTable)
	if err != nil {
		return loadedRoutingTable, customerrs.CouldNotMarshallError{Type: loadedRoutingTable, Err: err}
	}

	return loadedRoutingTable, nil
}

// GetAll returns all stored routingTables.
func (s *DatabaseRoutingTableStore) GetAll(ctx context.Context) (loadedRoutingTable []RoutingTable, err error) {
	cursor, err := s.collection.Find(ctx, bson.D{})
	if err != nil {
		return []RoutingTable{}, err
	}
	defer func() {
		if ferr := cursor.Close(ctx); ferr != nil {
			fErrString := ferr.Error()
			err = fmt.Errorf("InternalError=%w DeferError=%+s", err, fErrString)
		}
	}()

	err = cursor.All(ctx, &loadedRoutingTable)
	if err != nil {
		return loadedRoutingTable, customerrs.CouldNotMarshallError{Type: loadedRoutingTable, Err: err}
	}

	return loadedRoutingTable, nil
}

// Add adds a RoutingTable to the store.
func (s *DatabaseRoutingTableStore) Add(ctx context.Context, routingTable RoutingTable) (err error) {
	_, err = s.collection.
		InsertOne(ctx, routingTable)
	if err != nil {
		return customerrs.CouldNotCreateError{Identifier: routingTable.ID, Type: routingTable, Err: err}
	}

	return nil
}

// Update updates a existing routingTable.
func (s *DatabaseRoutingTableStore) Update(ctx context.Context, routingTable RoutingTable) (err error) {
	var updatedLoadedRoutingTable RoutingTable

	update := bson.D{primitive.E{Key: "$set", Value: routingTable}}

	upsert := false
	after := options.After
	opt := options.FindOneAndUpdateOptions{
		Upsert:         &upsert,
		ReturnDocument: &after,
	}

	err = s.collection.
		FindOneAndUpdate(
			ctx, bson.M{"_id": routingTable.ID.String()}, update, &opt).
		Decode(&updatedLoadedRoutingTable)
	if err != nil {
		return customerrs.CouldNotUpdateError{Identifier: routingTable.ID, Type: routingTable, Err: err}
	}

	return nil
}

// Delete deletes a node from the node store.
func (s *DatabaseRoutingTableStore) Delete(ctx context.Context, routingTable RoutingTable) (err error) {
	_, err = s.collection.DeleteOne(ctx, bson.D{primitive.E{Key: routingTable.ID.String()}})
	if err != nil {
		return err
	}

	return nil
}
